/* *      RTC Realtime Performance Test Application
*      Copyright (C) 2005, James Gibbons * *      Based on: *
*	Real Time Clock Driver Test/Example Program *
*	Copyright (C) 1996, Paul Gortmaker. *
*	Released under the GNU General Public License, version 2,
*	included herein by reference. * */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/resource.h>
#include <sched.h>
#include <sys/mman.h>

static bool ExitFlag = false;

// Ctrl-C hook function
void ctrlc( int sig )
{
  ExitFlag = true;
}

int main( int argc, char * argv[] )
{
  sched_param sp;
  int min_pr, max_pr;
  int EC;

  // lock all current and future memory in RAM
  EC = mlockall( MCL_CURRENT + MCL_FUTURE );
  if ( EC != 0 ) fprintf( stderr, "\n mlockall error %d ", EC );

  // get real-time priority min/max
  max_pr = sched_get_priority_max( SCHED_FIFO );
  min_pr = sched_get_priority_min( SCHED_FIFO );

  // bump us up to maximum priority
  sp.sched_priority = max_pr;
  EC = sched_setscheduler( 0, SCHED_FIFO, & sp );
  if ( EC != 0 ) fprintf( stderr, "\n sched_setscheduler error %d ", EC );

  // hook Ctrl-C signal
  struct sigaction act;

  act.sa_handler = ctrlc;
  sigemptyset( & act.sa_mask );
  act.sa_flags = 0;

  sigaction( SIGINT, & act, 0 );

  int RunTime = 60 * 60 * 8; // 8 hours

  // check to read optional run time parameter
  if ( argc > 1 )
  {
    errno = 0;
    char * tail;
    int num = strtol( argv[1], & tail, 0 );
    if ( ( errno == 0 ) && ( argv[1] != tail ) )
    {
      RunTime = num;
      if ( RunTime < 0 ) RunTime = 1;
    }
    else
    {
      fprintf( stderr, "\nParameter 1: time to run in seconds.\n", errno );
      return -1;
    }
  }

  int i, fd, retval;
  int irqcount = 0;
  unsigned long tmp, data;
  struct rtc_time rtc_tm;

  // open the RTC
  fd = open( "/dev/rtc", O_RDONLY );
  if ( fd == -1 )
  {
    perror( "/dev/rtc" );
    exit( errno );
  }

  // wipe screen and print startup message
  for ( i = 0; i < 80; i++ ) fprintf( stderr, "\n" );
  fprintf( stderr, "\nRTC Driver Real-Time Performance Test.\n" );
  fprintf( stderr, "\n *** Press Ctrl-C to exit early ***\n" );

  const int NMC = 100; // size of histogram
  unsigned long freq = 1024 * 4; // use 4K frequency
  double TimeBase = 1.0 / freq;
  unsigned int MissCount[NMC]; // histogram of interrupt counts
  unsigned int MaxMiss = 0; // maximum miss count

  for ( i = 0; i < NMC; i++ ) MissCount[i] = 0;

  // set the frequency
  retval = ioctl( fd, RTC_IRQP_SET, freq );
  if ( retval == -1 )
  {
    perror( "ioctl" );
    exit( errno );
  }

  fprintf( stderr, "\n\nRunTime: %d seconds at %ld Hz.\n", RunTime, freq );
  fflush( stderr );

  // Enable periodic interrupts
  retval = ioctl( fd, RTC_PIE_ON, 0 );
  if ( retval == -1 )
  {
    perror( "ioctl" );
    exit( errno );
  }

  // loop and read interrupts
  do
  {
    // This blocks until RTC interrupt
    retval = read( fd, & data, sizeof( unsigned long ) );
    if ( retval == -1 )
    {
      // check not Ctrl-C
      if ( errno != EINTR )
      {
        perror( "read" );
        exit( errno );
      }
    }
    else
    {
      // process histogram, mc is interrupt count
      int mc = ( data >> 8 ) & 0x00ffffff;

      irqcount += mc;

      if ( mc < 0 ) mc = 0;
      if ( mc >= NMC - 1 )
      {
        if ( mc > MaxMiss ) MaxMiss = mc;
        MissCount[NMC - 1] ++;
      }
      else
        MissCount[mc] ++;

      if ( !ExitFlag ) ExitFlag = irqcount >= freq * RunTime;
    }
  }
  while ( !ExitFlag );

  // Disable periodic interrupts
  retval = ioctl( fd, RTC_PIE_OFF, 0 );
  if ( retval == -1 )
  {
    perror( "ioctl" );
    exit( errno );
  }

  // close the RTC
  close( fd );

  // print the results
  fprintf( stderr, "\n\nInterrupt Count Histogram" );
  fprintf( stderr, "\n%4s %6s %10s %8s", "Ints", "MSec", "Total", "Percent" );

  int TotalMC = 0;

  for ( i = 0; i < NMC; i++ ) TotalMC += MissCount[i];

  for ( i = 0; i < NMC; i++ )
  {
    if ( MissCount[i] <= 0 ) continue;

    if ( i == NMC - 1 )
      fprintf( stderr, "\n%4s %6.1f", "Max", MaxMiss * TimeBase * 1000.0 );
    else
      fprintf( stderr, "\n%4d %6.1f", i, i * TimeBase * 1000.0 );

    fprintf( stderr, " %10d %8.3f", MissCount[i],
         100.0 * MissCount[i] / TotalMC );
  }

  fprintf( stderr, "\n\n *** Test complete ***\n" );

  return 0;
}
